@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1999, 2015 Tanuki Software, Ltd.
+ * Copyright (c) 1999, 2018 Tanuki Software, Ltd.
  * http://www.tanukisoftware.com
  * All rights reserved.
  *
@@ -20,6 +20,10 @@
 #else
  #include <glob.h>
  #include <string.h>
+ #include <limits.h>
+ #if defined(IRIX)
+  #define PATH_MAX FILENAME_MAX
+ #endif
 #endif
 
 #ifdef WIN32
@@ -49,6 +53,8 @@ int loggerFileGetSortMode(const TCHAR *modeName) {
         return LOGGER_FILE_SORT_MODE_NAMES_ASC;
     } else if (strcmpIgnoreCase(modeName, TEXT("NAMES_DEC")) == 0) {
         return LOGGER_FILE_SORT_MODE_NAMES_DEC;
+    } else if (strcmpIgnoreCase(modeName, TEXT("NAMES_SMART")) == 0) {
+        return LOGGER_FILE_SORT_MODE_NAMES_SMART;
     } else {
         return LOGGER_FILE_SORT_MODE_TIMES;
     }
@@ -190,6 +196,42 @@ int compareFileNames(const TCHAR *file1, const TCHAR *file2) {
     }
 }
 
+int compareFileNamesIndex(const TCHAR *file1, const TCHAR *file2, int startIndex, int stopIndex, int startCountFromEnd, int stopCountFromEnd) {
+    TCHAR *file1_;
+    TCHAR *file2_;
+    int start;
+    int len;
+    int result;
+    
+    start = (startCountFromEnd ? (int)_tcslen(file1) - startIndex : startIndex);
+    len   = (stopCountFromEnd  ? (int)_tcslen(file1) - stopIndex  : stopIndex) - start;
+    file1_ = malloc(sizeof(TCHAR) * (len + 1));
+    if (!file1_) {
+        outOfMemoryQueued(TEXT("CFNI"), 1);
+        return 0;
+    }
+    _tcsncpy(file1_, file1 + start, len);
+    file1_[len] = 0;
+    
+    start = (startCountFromEnd ? (int)_tcslen(file2) - startIndex : startIndex);
+    len   = (stopCountFromEnd  ? (int)_tcslen(file2) - stopIndex  : stopIndex) - start;
+    file2_ = malloc(sizeof(TCHAR) * (len + 1));
+    if (!file2_) {
+        free(file1_);
+        outOfMemoryQueued(TEXT("CFNI"), 2);
+        return 0;
+    }
+    _tcsncpy(file2_, file2 + start, len);
+    file2_[len] = 0;
+    
+    result = compareFileNames(file1_, file2_);
+    
+    free(file1_);
+    free(file2_);
+    
+    return result;
+}
+
 int sortFilesNamesAsc(TCHAR **files, int cnt) {
     int i, j;
     TCHAR *temp;
@@ -228,6 +270,94 @@ int sortFilesNamesDec(TCHAR **files, int cnt) {
     return TRUE;
 }
 
+int sortFilesNamesDecIndex(TCHAR **files, int cnt, int startIndex, int stopIndex, int startCountFromEnd, int stopCountFromEnd) {
+    int i, j;
+    TCHAR *temp;
+    int cmp;
+
+    for (i = 0; i < cnt; i++) {
+        for (j = 0; j < cnt - 1; j++) {
+            cmp = compareFileNamesIndex(files[j], files[j+1], startIndex, stopIndex, startCountFromEnd, stopCountFromEnd);
+            if (cmp > 0) {
+                temp = files[j + 1];
+                files[j + 1] = files[j];
+                files[j] = temp;
+            }
+        }
+    }
+
+    return TRUE;
+}
+
+/**
+ * This function allows to sort filenames with the following logic:
+ *  - if the given pattern contains a ???????? (date) token, the files are first sorted by date descending.
+ *  - if the given pattern contains a * (num) token, the files are sorted by ascending numbers (for equivalent dates).
+ *     The first file being the one without number.
+ *  NOTE: The function assumes that there is at most one token of each!
+ *
+ * @param pattern pattern used to figure out how to sort the files.
+ * @param files list of files to sort.
+ * @param cnt number of files to sort.
+ */
+int sortFilesNamesSmart(const TCHAR* pattern, TCHAR **files, int cnt) {
+    int i, j;
+    TCHAR *temp;
+    int cmp;
+    TCHAR* numToken;
+    TCHAR* dateToken;
+    int numStartIndex, numStopIndex;
+    int dateStartIndex = 0;
+    int dateStopIndex = 0;
+    int startCountFromEnd = 0;
+    int stopCountFromEnd = 0;
+    
+    /* First sort by date. */
+    dateToken = _tcsstr(pattern, TEXT("?"));
+    numToken = _tcsstr(pattern, TEXT("*"));
+    
+    if (dateToken) {
+        if (!numToken || (dateToken < numToken)) {
+            dateStartIndex = (int)(dateToken - pattern);
+            dateStopIndex = dateStartIndex + 8;
+            startCountFromEnd = FALSE;
+            stopCountFromEnd = FALSE;
+        } else {
+            /* There is a num token before the date. So the length before the date is not fixed. Calculate the index from the end. */
+            dateStartIndex = (int)_tcslen(pattern) - (int)(dateToken - pattern);
+            dateStopIndex = dateStartIndex - 8;
+            startCountFromEnd = TRUE;
+            stopCountFromEnd = TRUE;
+        }
+        sortFilesNamesDecIndex(files, cnt, dateStartIndex, dateStopIndex, startCountFromEnd, stopCountFromEnd);
+    }
+    
+    if (numToken) {
+        numStartIndex = (int)(numToken - pattern);
+        numStopIndex = (int)_tcslen(pattern) - (numStartIndex + 1);
+        for (i = 0; i < cnt; i++) {
+            for (j = 0; j < cnt - 1; j++) {
+                if (dateToken) {
+                    /* Make sure that the dates are equals. */
+                    cmp = compareFileNamesIndex(files[j], files[j+1], dateStartIndex, dateStopIndex, startCountFromEnd, stopCountFromEnd);
+                    if (cmp != 0) {
+                        continue;
+                    }
+                }
+                /* Sort by ascending name. */
+                cmp = compareFileNamesIndex(files[j], files[j+1], numStartIndex, numStopIndex, FALSE, TRUE);
+                if (cmp < 0) {
+                    temp = files[j + 1];
+                    files[j + 1] = files[j];
+                    files[j] = temp;
+                }
+            }
+        }
+    }
+
+    return TRUE;
+}
+
 /**
  * Returns a NULL terminated list of file names within the specified pattern.
  *  The files will be sorted new to old for TIMES.  Then incremental ordering
@@ -547,6 +677,13 @@ TCHAR** loggerFileGetFiles(const TCHAR* pattern, int sortMode) {
             loggerFileFreeFiles(files);
             return NULL;
         }
+    } else if (sortMode == LOGGER_FILE_SORT_MODE_NAMES_SMART) {
+        if (!sortFilesNamesSmart(pattern, files, cnt)) {
+            /* Failed. Reported. */
+            free(fileTimes);
+            loggerFileFreeFiles(files);
+            return NULL;
+        }
     } else if (sortMode == LOGGER_FILE_SORT_MODE_NAMES_DEC) {
         if (!sortFilesNamesDec(files, cnt)) {
             /* Failed. Reported. */
@@ -590,6 +727,247 @@ void loggerFileFreeFiles(TCHAR** files) {
     free(files);
 }
 
+/**
+ * Combines two paths and take care to add only one separator between them.
+ *
+ * The returned string must be freed by the caller.
+ *
+ * @param path1 base path
+ * @param path2 relative path.
+ *
+ * @return The resulting path, or NULL if there were any problems.
+ */
+TCHAR *combinePath(const TCHAR *path1, const TCHAR *path2) {
+    TCHAR* result;
+    TCHAR* tempPath1 = NULL;
+    TCHAR* tempPath2 = NULL;
+    TCHAR* tempPath2Ptr;
+    TCHAR c1;
+    TCHAR c2;
+    size_t len1 = _tcslen(path1); 
+    size_t len2 = _tcslen(path2); 
+    size_t len = len1 + len2;
+    int i = 0;
+#ifdef WIN32
+    const TCHAR bad_sep  = TEXT('/');
+    const TCHAR good_sep = TEXT('\\');
+#else
+    const TCHAR bad_sep  = TEXT('\\');
+    const TCHAR good_sep = TEXT('/');
+#endif
+
+    if (len1 > 0) {
+        tempPath1 = malloc(sizeof(TCHAR) * (len1 + 1));
+        if (!tempPath1) {
+            outOfMemoryQueued(TEXT("CP"), 1);
+            return NULL;
+        }
+        _tcsncpy(tempPath1, path1, len1 + 1);
+    }
+    if (len2 > 0) {
+        tempPath2 = malloc(sizeof(TCHAR) * (len2 + 1));
+        if (!tempPath2) {
+            outOfMemoryQueued(TEXT("CP"), 2);
+            free(tempPath1);
+            return NULL;
+        }
+        _tcsncpy(tempPath2, path2, len2 + 1);
+    }
+    if (!tempPath1 && !tempPath2) {
+        result = NULL;
+    } else if (tempPath1 && !tempPath2) {
+        result = tempPath1;
+    } else if (!tempPath1 && tempPath2) {
+        result = tempPath2;
+    } else {
+        tempPath2Ptr = tempPath2;
+
+        /* first replace all directory separators by their standard according to the platform. 
+         *  we want to avoid that the two paths use different separators. */
+        while (tempPath1[i] != TEXT('\0')) {
+            if (tempPath1[i] == bad_sep) {
+                tempPath1[i] = good_sep;
+            }
+            i++;
+        }
+        i = 0;
+        while (tempPath2[i] != TEXT('\0')) {
+            if (tempPath2[i] == bad_sep) {
+                tempPath2[i] = good_sep;
+            }
+            i++;
+        }
+        
+        c1 = tempPath1[len1 - 1];
+        c2 = tempPath2[0];
+        
+        if (c1 == good_sep) {
+            if (c2 == good_sep) {
+                tempPath2Ptr++;
+            } else {
+                len += 1;
+            }
+            
+            result = malloc(sizeof(TCHAR) * len);
+            if (!result) {
+                outOfMemoryQueued(TEXT("CP"), 3);
+                free(tempPath1);
+                free(tempPath2);
+                return NULL;
+            }
+            _sntprintf(result, len, TEXT("%s%s"), tempPath1, tempPath2Ptr);
+        } else {
+            if (c2 == good_sep) {
+                tempPath2Ptr++;
+                len += 1;
+            } else {
+                len += 2;
+            }
+            
+            result = malloc(sizeof(TCHAR) * len);
+            if (!result) {
+                outOfMemoryQueued(TEXT("CP"), 4);
+                free(tempPath1);
+                free(tempPath2);
+                return NULL;
+            }
+            _sntprintf(result, len, TEXT("%s%c%s"), tempPath1, good_sep, tempPath2Ptr);
+        }
+        free(tempPath1);
+        free(tempPath2);
+    }
+    return result;
+}
+
+/**
+ * Given a path, resolve a real absolute path which has resolved all relative and symbolic links.
+ *
+ * The returned string must be freed by the caller.
+ *
+ * @param path The source path.
+ * @param pathDesc A description of the path used for error messages.
+ * @param errorLevel Level to log errors at.
+ *
+ * @return The absolute path, or NULL if there were any problems.
+ */
+TCHAR *getRealPath(const TCHAR *path, const TCHAR *pathDesc, int errorLevel, int useQueue) {
+    TCHAR *realPath;
+#ifdef WIN32
+    DWORD len;
+#else
+    size_t len;
+    TCHAR *tempPath;
+#endif
+    
+#ifdef WIN32
+    len = GetFullPathName(path, 0, NULL, NULL);
+    if (!len) {
+        if (errorLevel != LEVEL_NONE) {
+            log_printf_queue(useQueue, WRAPPER_SOURCE_WRAPPER, errorLevel, TEXT("Unable to resolve the %s '%s': %s"), pathDesc, path, getLastErrorText());
+        }
+        return NULL;
+    }
+    realPath = malloc(sizeof(TCHAR) * len);
+    if (!realPath) {
+        if (useQueue) {
+            outOfMemoryQueued(TEXT("GRP"), 1);
+        } else {
+            outOfMemory(TEXT("GRP"), 1);
+        }
+        return NULL;
+    }
+    if (!GetFullPathName(path, len, realPath, NULL)) {
+        if (errorLevel != LEVEL_NONE) {
+            log_printf_queue(useQueue, WRAPPER_SOURCE_WRAPPER, errorLevel, TEXT("Unable to resolve the %s '%s': %s"), pathDesc, path, getLastErrorText());
+        }
+        free(realPath);
+        return NULL;
+    }
+#else
+    /* The solaris implementation of realpath will return a relative path if a relative
+     *  path is provided.  We always need an absolute path here.  So build up one and
+     *  then use realpath to remove any .. or other relative references. */
+    tempPath = malloc(sizeof(TCHAR) * (PATH_MAX + 1));
+    if (!tempPath) {
+        if (useQueue) {
+            outOfMemoryQueued(TEXT("GRP"), 2);
+        } else {
+            outOfMemory(TEXT("GRP"), 2);
+        }
+        return NULL;
+    }
+    if (_trealpathN(path, tempPath, PATH_MAX + 1) == NULL) {
+        if (errorLevel != LEVEL_NONE) {
+            log_printf_queue(useQueue, WRAPPER_SOURCE_WRAPPER, errorLevel, TEXT("Unable to resolve the %s '%s': %s"), pathDesc, path, getLastErrorText());
+        }
+        free(tempPath);
+        return NULL;
+    }
+    
+    /* Now that we know how big the resulting string is, put it into a buffer of the correct size to avoid waste. */
+    len = _tcslen(tempPath) + 1;
+    realPath = malloc(sizeof(TCHAR) * len);
+    if (!realPath) {
+        if (useQueue) {
+            outOfMemoryQueued(TEXT("GRP"), 3);
+        } else {
+            outOfMemory(TEXT("GRP"), 3);
+        }
+        free(tempPath);
+        return NULL;
+    }
+    _tcsncpy(realPath, tempPath, len);
+    free(tempPath);
+#endif
+    
+    return realPath;
+}
+
+/**
+ * Returns the absolute path of a file even if the file is not yet created.
+ *  The folder containing the file must exist.
+ *
+ * The returned string must be freed by the caller.
+ *
+ * @param path The source path.
+ * @param pathDesc A description of the path used for error messages.
+ * @param errorLevel Level to log errors at.
+ *
+ * @return The absolute path, or NULL if there were any problems.
+ */
+TCHAR* getAbsolutePathOfFile(const TCHAR* path, const TCHAR *pathDesc, int errorLevel, int useQueue) {
+    TCHAR* ptr;
+    TCHAR* dir;
+    const TCHAR* file;
+    TCHAR* result = NULL;
+    TCHAR* pathCpy;
+    
+    pathCpy = malloc(sizeof(TCHAR) * (_tcslen(path) + 1));
+    if (!pathCpy) {
+        outOfMemoryQueued(TEXT("GAPOF"), 1);
+    } else {
+        _tcsncpy(pathCpy, path, _tcslen(path) + 1);
+        ptr = __max(_tcsrchr(pathCpy, TEXT('\\')), _tcsrchr(pathCpy, TEXT('/')));
+        if (ptr) {
+            *ptr = 0;
+            ptr++;
+            dir = getRealPath(pathCpy, pathDesc, errorLevel, useQueue);
+            file = ptr;
+        } else {
+            dir = getRealPath(TEXT("."), pathDesc, errorLevel, useQueue);
+            file = pathCpy;
+        }
+        
+        if (dir) {
+            result = combinePath(dir, file);
+            free(dir);
+        }
+        free(pathCpy);
+    }
+    
+    return result;
+}
+
 #ifdef LOGGER_FILE_DEBUG
 void loggerFileTests() {
     TCHAR** files;
